(* $Id: taskalg.ml,v 1.1 2010-07-23 15:00:31 deraugla Exp $ *)

open Tasks;
open Printf;

(*
module Sarray = Array;
type sarray_t 'a = array 'a;
value sarray_empty = [| |];
*)
type sarray_t 'a = Sarray.t 'a;
value sarray_empty = Sarray.empty;
(**)

type t =
  { st_release : int;
    st_init : sarray_t (option time_unit);
    st_repeat : sarray_t (option time_unit);
    st_non_preemptive : bool;
    st_contents : list task_component }
and time_unit =
  { tu_itask : int;
    tu_is_first : bool;
    tu_is_preemption : bool }
and task_component =
  { tc_itask : int;
    tc_release : int;
    tc_duration : int;
    tc_period : int }
;

type stask_bucket =
  { sb_task : t;
    sb_pos : int;
    sb_len : int }
;

value string_of_sarray string_of_elem arr =
  let list = List.map string_of_elem (Sarray.to_list arr) in
  String.concat "" list
;

value col = [| 2; 1; 4; 5; 6; 3 |];
value col tu = do {
  let n = tu.tu_itask in
  if n < 0 then 0 else col.(n mod Array.length col)
};

value string_of_option_unit =
  fun
  [ Some tu -> do {
      sprintf "\027[%sm%d\027[m"
        (if tu.tu_is_preemption then "7" else sprintf "1;3%d" (col tu))
        (tu.tu_itask + 1)
    }
  | None -> "." ]
;

value string_of_stask_area = string_of_sarray string_of_option_unit;

value to_string st = do {
  sprintf "%s{%s} (rel %d)" (string_of_stask_area st.st_init)
    (string_of_stask_area st.st_repeat) st.st_release
};

value repeat_of_task_def i td =
  loop [] td.duration td.period
  where rec loop rev_list dur per =
    if per = 0 then Sarray.of_list (List.rev rev_list)
    else if dur = 0 then loop [None :: rev_list] 0 (per - 1)
    else do {
      let tu =
        let is_first = rev_list = [] in
        {tu_itask = i; tu_is_first = is_first; tu_is_preemption = False}
      in
      loop [Some tu :: rev_list] (dur - 1) (per - 1)
    }
;

value of_task_def i td = do {
  if td.sub_duration <> [] then
    failwith (sprintf "sub duration not impl (t%d)" (i + 1))
  else ();
  let tc =
    {tc_itask = i; tc_release = td.release; tc_duration = td.duration;
     tc_period = td.period}
  in
  {st_release = td.release;
   st_init = sarray_empty;
   st_repeat = repeat_of_task_def i td;
   st_non_preemptive = td.non_preemptive;
   st_contents = [tc]}
};

value stask_bucket_of_stask st =
  {sb_task = st; sb_pos = 0; sb_len = Sarray.length st.st_init}
;

value stask_next sb = do {
  if sb.sb_pos >= sb.sb_len then
    None
  else
    let v = Sarray.get sb.sb_task.st_init sb.sb_pos in
    let sb = {(sb) with sb_pos = sb.sb_pos + 1} in
    Some (v, sb)
};

value copy_repeat_on_init sb =
  let st = {(sb.sb_task) with st_init = sb.sb_task.st_repeat} in
  {sb_task = st; sb_pos = 0; sb_len = Sarray.length st.st_repeat}
;

value rec stask_bucket_next sb = do {
  match stask_next sb with
  [ Some x -> x
  | None -> stask_bucket_next (copy_repeat_on_init sb) ]
};

value update_release tasks_prec stask1 stask2 tc = do {
  match tasks_prec with
  [ [] -> Some stask2
  | pt -> do {
      let release_opt =
        loop_prec stask2.st_release pt where rec loop_prec release =
          fun
          [ [itask :: itaskl] -> do {
               loop stask1.st_contents where rec loop =
                 fun
                 [ [tc :: tcl] -> do {
                     if tc.tc_itask = itask then do {
                       let d = tc.tc_duration in
                       let sb = stask_bucket_of_stask stask1 in
                       loop stask1.st_release d (stask_bucket_next sb)
                       where rec loop rel d (v, sb) =
                         if d = 0 then loop_prec (max rel release) itaskl
                         else do {
                           let d =
                             match v with
                             [ Some tu ->
                                 if tu.tu_itask = itask &&
                                    not tu.tu_is_preemption
                                 then
                                   d - 1
                                 else d
                             | None -> d ]
                           in
                           loop (rel + 1) d (stask_bucket_next sb)
                         }
                     }
                     else loop tcl
                   }
                 | [] -> None ]
            }
          | [] -> Some release ]
      in
      match release_opt with
      [ Some release -> do {
          let tc = {(tc) with tc_release = release} in
          Some {(stask2) with st_release = release; st_contents = [tc]}
        }
      | None -> None ]
    } ]
};

value adjust_release diff_rel sb =
  let st =
    let st = sb.sb_task in
    {(st) with
     st_release = st.st_release - diff_rel;
     st_init = Sarray.append (Sarray.make diff_rel None) st.st_init}
  in
  {sb_task = st; sb_pos = 0; sb_len = Sarray.length st.st_init}
;

value adjust_releases sb1 sb2 =
  let st1 = sb1.sb_task in
  let st2 = sb2.sb_task in
  if st1.st_release < st2.st_release then do {
    let n = st2.st_release - st1.st_release in
    let sb2 = adjust_release n sb2 in
    (sb1, sb2)
  }
  else if st1.st_release > st2.st_release then do {
    let n = st1.st_release - st2.st_release in
    let sb1 = adjust_release n sb1 in
    (sb1, sb2)
  }
  else (sb1, sb2)
;

module type S =
  sig
    type t 'a = 'abstract;
    value empty : t 'a;
    value add : 'a -> t 'a -> t 'a;
    value get_init_and_repeat :
      int -> t 'a -> option (sarray_t 'a * sarray_t 'a);
    value to_string : ('a -> string) -> t 'a -> string;
  end
;

module S1 : S =
  struct
    type t 'a = list 'a;
    value empty = [];
    value add x t = [x :: t];
    value list_rev_first_elem n list =
      loop n [] list where rec loop n r list =
        if n = 0 then (r, list)
        else do {
          match list with
          [ [x :: l] -> loop (n - 1) [x :: r] l
          | [] -> (r, []) ]
        }
    ;
    value get_init_and_repeat n t = do {
      let (list1, rev_rest) = list_rev_first_elem n t in
      let (list2, rev_rest) = list_rev_first_elem n rev_rest in
      if list1 = list2 then do {
        let (rev_init, repeat_hd, rev_repeat) =
          loop rev_rest [] (List.rev list1)
          where rec loop rev_init repeat_hd rev_repeat =
            match (rev_init, rev_repeat) with
            [ ([x :: l1], [y :: l2]) -> do {
                if x = y then loop l1 [y :: repeat_hd] l2
                else (rev_init, repeat_hd, rev_repeat)
              }
            | (_, _) -> (rev_init, repeat_hd, rev_repeat) ]
        in
        let init = Sarray.of_list (List.rev rev_init) in
        let repeat = Sarray.of_list (repeat_hd @ (List.rev rev_repeat)) in
        Some (init, repeat)
      }
      else None
    };
    value to_string string_of_elem list =
      let list = List.rev_map string_of_elem list in
      String.concat "" list
    ;
  end
;

module S2 : S =
  struct
    type t 'a = {ea_arr : sarray_t 'a; ea_pos : int};
    value empty = {ea_arr = sarray_empty; ea_pos = 0};
    value add x ea = do {
      let len = Sarray.length ea.ea_arr in
      let a =
        if ea.ea_pos = len then do {
          let new_len = 2 * len + 1 in
          Sarray.init new_len
            (fun i -> if i < len then Sarray.get ea.ea_arr i else x)
        }
        else do {
          Sarray.set ea.ea_arr ea.ea_pos x;
          ea.ea_arr
        }
      in
      {ea_arr = a; ea_pos = ea.ea_pos + 1}
    };
    value get_init_and_repeat n ea = do {
      let init_len =
        loop n (ea.ea_pos - 1) (ea.ea_pos - n - 1) where rec loop k i j =
          if j < 0 then 0
          else if Sarray.get ea.ea_arr i = Sarray.get ea.ea_arr j then
            loop (k - 1) (i - 1) (j - 1)
          else if k < 0 then j + 1
          else -1
      in
      if init_len >= 0 then do {
        let init = Sarray.sub ea.ea_arr 0 init_len in
        let repeat = Sarray.sub ea.ea_arr init_len n in
        Some (init, repeat)
      }
      else None
    };
    value to_string string_of_elem ea =
      let a = Sarray.sub ea.ea_arr 0 ea.ea_pos in
      let list = List.map string_of_elem (Sarray.to_list a) in
      String.concat "" list
    ;
  end
;

module S = S2;

value check_repeat per_lcm accu sb1 sb2 = do {
  match S.get_init_and_repeat per_lcm accu with
  [ Some (init, repeat) -> do {
      let contents = sb1.sb_task.st_contents @ sb2.sb_task.st_contents in
      let st =
        {(sb1.sb_task) with
         st_init = init; st_repeat = repeat; st_contents = contents}
      in
      Some st
    }
  | None -> None ]
};

value shorten_end_if_free sb = do {
  if sb.sb_len = 0 then None
  else do {
    match Sarray.get sb.sb_task.st_init (sb.sb_len - 1) with
    [ None -> do {
        let sb = {(sb) with sb_pos = sb.sb_pos - 1; sb_len = sb.sb_len - 1} in
        Some sb
      }
    | Some _ -> None ]
  }
};

value select_unit tcon preemption v1 v2 sb1 sb2 = do {
  match (v1, v2) with
  [ (Some _, Some tu2) -> do {
      if sb2.sb_pos >= sb2.sb_len then None
      else do {
        match shorten_end_if_free sb2 with
        [ Some sb2 -> do {
            let preemption =
              if not tu2.tu_is_first then tcon.preemption_cost else 0
            in
            Some (sb1, sb2, v1, preemption)
          }
        | None -> None ]
      }
    }
  | (None, Some tu2) -> do {
      if sb2.sb_task.st_non_preemptive && tu2.tu_is_first then do {
        (* first time unit in non preemptive task: check if room enough;
         * put nothing (= None) if not. *)
        let dur =
          loop 1 sb2 where rec loop dur sb2 =
            match stask_next sb2 with
            [ Some (Some tu, sb2) -> do {
                if tu.tu_itask = tu2.tu_itask then loop (dur + 1) sb2
                else dur
              }
            | Some (None, _) | None -> dur ]
        in
        loop (dur - 1) sb1 where rec loop dur_rest sb =
          if dur_rest = 0 then Some (sb1, sb2, v2, 0)
          else do {
            let (v, sb) = stask_bucket_next sb in
            match v with
            [ Some _ -> do {
                match shorten_end_if_free sb2 with
                [ Some sb2 -> Some (sb1, sb2, None, 0)
                | None -> None ]
              }
            | None -> loop (dur_rest - 1) sb ]
          }
      }
      else if preemption > 0 then do {
        match Sarray.get sb2.sb_task.st_init (sb2.sb_len - 1) with
        [ None -> do {
            let sb2 =
              {(sb2) with sb_pos = sb2.sb_pos - 1; sb_len = sb2.sb_len - 1}
            in
            let tu =
              {tu_itask = tu2.tu_itask; tu_is_first = False;
               tu_is_preemption = True}
            in
            Some (sb1, sb2, Some tu, preemption - 1)
          }
        | Some _ -> None ]
      }
      else Some (sb1, sb2, v2, 0)
    }
  | (Some _, None) -> Some (sb1, sb2, v1, 0)
  | (None, None) -> Some (sb1, sb2, None, 0) ]
};

value plus_loop tcon per_lcm = do {
  loop S.empty 0 where rec loop accu p n sb1 sb2 = do {
    match if n = 0 then check_repeat per_lcm accu sb1 sb2 else None with
    [ Some r -> Some r
    | None -> do {
        let n = if n = 0 then per_lcm else n in
        let (v1, sb1) = stask_bucket_next sb1 in
        let (v2, sb2) = stask_bucket_next sb2 in
        match select_unit tcon p v1 v2 sb1 sb2 with
        [ Some (sb1, sb2, v, p) -> loop (S.add v accu) p (n - 1) sb1 sb2
        | None -> None ]
      } ]
  }
};

value plus tcon st1 st2 = do {
  let st2_opt =
    match st2.st_contents with
    [ [tc] -> do {
        let tasks_prec =
          Sparse.graph_preceding_tasks tcon.graph_precedence tc.tc_itask
        in
        let tasks_prec =
          if tasks_prec = [] && tcon.global_precedence then do {
            List.map (fun tc1 -> tc1.tc_itask) st1.st_contents
          }
          else tasks_prec
        in
        update_release tasks_prec st1 st2 tc
      }
    | [] | [_ :: _] -> Some st2 ]
  in
  match st2_opt with
  [ Some st2 -> do {
      let per_lcm =
        Sutil.lcm (Sarray.length st1.st_repeat) (Sarray.length st2.st_repeat)
      in
      let sb1 = stask_bucket_of_stask st1 in
      let sb2 = stask_bucket_of_stask st2 in
      let (sb1, sb2) = adjust_releases sb1 sb2 in
      let rmax =
        max (Sarray.length st1.st_init) (Sarray.length st2.st_init)
      in
      let min_sched = rmax + 2 * per_lcm in
      plus_loop tcon per_lcm min_sched sb1 sb2
    }
  | None -> None ]
};

value response_time_of_task st =
  let a =
    Sarray.append (Sarray.append st.st_init st.st_repeat) st.st_repeat
  in
  fun only_first tc ->
    loop 0 (tc.tc_release - st.st_release) where rec loop wrt rel =
      loop_inst rel tc.tc_duration where rec loop_inst i dur =
        if dur = 0 then do {
          let wrt = max wrt (i - rel) in
          if only_first then wrt else loop wrt (rel + tc.tc_period)
        }
        else if i >= Sarray.length a then wrt
        else do {
          let dur =
            match Sarray.get a i with
            [ Some tu -> do {
                if tu.tu_itask = tc.tc_itask && not tu.tu_is_preemption then
                  dur - 1
                else dur
              }
            | None -> dur ]
          in
          loop_inst (i + 1) dur
        }
;

value first_response_time st = do {
  let f = response_time_of_task st in
  let wrt = List.map (fun tc -> (tc.tc_itask, f True tc)) st.st_contents in
  Array.of_list wrt
};

value worst_response_time st = do {
  let f = response_time_of_task st in
  let wrt = List.map (fun tc -> (tc.tc_itask, f False tc)) st.st_contents in
  Array.of_list wrt
};
