(* $Id: tasks.mli,v 1.1 2010-07-23 15:00:31 deraugla Exp $ *)

type tasks =
  { task : mutable array task_def;
    task_order : mutable array int;
    priority_policy : mutable  priority_policy;
    user_graph_precedence : mutable list (list int);
    tasks_constraints : mutable tasks_constraints;
    strict_periodicity : mutable bool;
    fully_non_preemptive : mutable bool;
    latency_constr : mutable list latency_constr;

    schedule_len : mutable int;
    disp_schedule : mutable Sarray.t (option time_unit);
    comp_schedule : mutable Sarray.t (option time_unit);
    rmax : mutable int;
    lcm : mutable int;
    min_rel_time : mutable int;

    schedule_status : mutable schedule_status;
    repetition : mutable option (int * int);
    curr_ord_ind : mutable int;
    curr_inst : mutable int;
    do_refresh : mutable bool }
and task_def =
  { release : int; duration : int; sub_duration : list sub_duration;
    deadline : int; period : int; user_release : int; non_preemptive : bool;
    strict_per : bool; pree_cost : int }
and tasks_constraints = 
  { global_precedence : mutable bool;
    graph_precedence : mutable list (list int);
    preemption_cost : mutable int }
and sub_duration =
  { sub_rel : int; sub_dur : int; sub_non_pree : bool; sub_str_per : bool }
and time_unit =
  { u_task : int; u_inst : int; u_preempt : bool }
and latency_constr =
  { l_task1 : int; l_inst1 : int; l_task2 : int; l_inst2 : int; l_max : int;
    l_status : mutable latency_status }
and priority_policy =
  [ PP_deadline_monotonic
  | PP_rate_monotonic
  | PP_edf
  | PP_audsley of audsley_info
  | PP_user of array int ]
and audsley_info =
  { ai_pos : position;
    ai_tab : array int;
    ai_step : int;
    ai_nsol : int }
and position =
  [ P_begin
  | P_middle
  | P_end ]
and schedule_status =
  [ SS_running
  | SS_schedulable_but of string
  | SS_task_error of int and string
  | SS_incomplete_scheduling of int and string
  | SS_several_lcm of int ]
and latency_status =
  [ LS_unchecked
  | LS_out_of_bounds
  | LS_invalid
  | LS_values of (int * int) and (int * int) and int ]
;

type alt 'a 'b = [ Left of 'a | Right of 'b ];
